# Copyright (c) 2013 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from zaqarclient import errors


def version(min_version, max_version=None):
    min_version = float(min_version)
    max_version = max_version and float(max_version)

    error_msg = ('Method %(name)s is supported from version %(min)s '
                 'to %(max)s')

    def method(meth):
        @functools.wraps(meth)
        def wrapper(self, *args, **kwargs):
            if (self.api_version < min_version or
                    (max_version and self.api_version > max_version)):
                msg = error_msg % dict(name=str(meth),
                                       min=min_version,
                                       max=max_version or 'latest')
                raise errors.UnsupportedVersion(msg)
            return meth(self, *args, **kwargs)
        return wrapper
    return method


def lazy_property(write=False, delete=True):
    """Creates a lazy property.

    :param write: Whether this property is "writable"
    :param delete: Whether this property can be deleted.
    """

    def wrapper(fn):
        attr_name = '_lazy_' + fn.__name__

        def getter(self):
            if not hasattr(self, attr_name):
                setattr(self, attr_name, fn(self))
            return getattr(self, attr_name)

        def setter(self, value):
            setattr(self, attr_name, value)

        def deleter(self):
            delattr(self, attr_name)

        return property(fget=getter,
                        fset=write and setter,
                        fdel=delete and deleter,
                        doc=fn.__doc__)
    return wrapper



# Copyright (c) 2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from zaqarclient.queues.v1 import claim
from zaqarclient.queues.v1 import iterator as iterate
from zaqarclient.queues.v2 import core
from zaqarclient.queues.v2 import message


class Claim(claim.Claim):

    def _create(self):
        req, trans = self._queue.client._request_and_transport()
        msgs = core.claim_create(trans, req,
                                 self._queue._name,
                                 ttl=self._ttl,
                                 grace=self._grace,
                                 limit=self._limit)

        # extract the id from the first message
        if msgs is not None:
            if self._queue.client.api_version >= 1.1:
                msgs = msgs['messages']
            self.id = msgs[0]['href'].split('=')[-1]

        self._message_iter = iterate._Iterator(self._queue.client,
                                               msgs or [],
                                               'messages',
                                               message.create_object(
                                                   self._queue
                                               ))



# Copyright (c) 2013 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

import six

from zaqarclient.transport import errors


@six.add_metaclass(abc.ABCMeta)
class Transport(object):

    # common HTTP codes used by multiple transports
    http_to_zaqar = {
        400: errors.MalformedRequest,
        401: errors.UnauthorizedError,
        403: errors.ForbiddenError,
        404: errors.ResourceNotFound,
        409: errors.ConflictError,
        500: errors.InternalServerError,
        503: errors.ServiceUnavailableError
    }

    def __init__(self, options):
        self.options = options

    @abc.abstractmethod
    def send(self, request):
        """Returns the response.

        :returns: The final response
        :rtype: `zaqarclient.transport.response.Response`
        """



# Copyright (c) 2013 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from zaqarclient.auth import base
from zaqarclient.auth import keystone
from zaqarclient.auth import signed_url

_BACKENDS = {
    'noauth': base.NoAuth,
    'keystone': keystone.KeystoneAuth,
    'signed-url': signed_url.SignedURLAuth,
}


def get_backend(backend='keystone', options=None):
    """Loads backend `auth_backend`

    :params backend: The backend name to load.
        Default: `keystone`
    :type backend: `six.string_types`
    :param options: Options to pass to the Auth
        backend. Refer to the backend for more info.
    :type options: `dict`
    """
    if options is None:
        options = {}

    backend = _BACKENDS[backend](options)
    return backend



# Copyright (c) 2013  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import fixtures
import openstack.config
from six.moves import urllib_parse
import testtools

_USE_AUTHENTICATION = os.environ.get('ZAQARCLIENT_AUTH_FUNCTIONAL', False)
_RUN_FUNCTIONAL = os.environ.get('ZAQARCLIENT_TEST_FUNCTIONAL',
                                 _USE_AUTHENTICATION)


class TestBase(testtools.TestCase):

    transport_cls = None
    is_functional = False

    def setUp(self):
        super(TestBase, self).setUp()

        self.conf = {
            'auth_opts': {
                'backend': 'noauth',
                'options': {
                    'os_project_id': 'my-project'
                }
            }
        }
        self.useFixture(fixtures.FakeLogger('zaqar'))

        # NOTE(kgriffs): Don't monkey-patch stdout since it breaks
        # debugging with pdb.
        stderr = self.useFixture(fixtures.StringStream('stderr')).stream
        self.useFixture(fixtures.MonkeyPatch('sys.stderr', stderr))

        if not _RUN_FUNCTIONAL and self.is_functional:
            self.skipTest('Functional tests disabled')
        elif self.is_functional and _USE_AUTHENTICATION:
            self._setup_auth_params()

    def _credentials(self, cloud='devstack-admin'):
        """Retrieves credentials to run functional tests

        Credentials are either read via os-client-config from the environment
        or from a config file ('clouds.yaml'). Environment variables override
        those from the config file.

        devstack produces a clouds.yaml with two named clouds - one named
        'devstack' which has user privs and one named 'devstack-admin' which
        has admin privs. This function will default to getting the
        devstack-admin cloud as that is the current expected behavior.
        """
        os_cfg = openstack.config.OpenStackConfig()
        try:
            found = os_cfg.get_one_cloud(cloud=cloud)
        except Exception:
            found = os_cfg.get_one_cloud()
        return found

    def _setup_auth_params(self):
        self.creds = self._credentials().get_auth_args()

        parsed_url = urllib_parse.urlparse(self.creds['auth_url'])
        auth_url = self.creds['auth_url']
        if not parsed_url.path or parsed_url.path == '/':
            auth_url = urllib_parse.urljoin(self.creds['auth_url'], 'v3')
        if parsed_url.path == '/identity':
            auth_url = '%s/v3' % auth_url

        self.conf['auth_opts']['backend'] = 'keystone'
        options = {'os_username': self.creds['username'],
                   'os_user_domain_id': self.creds['user_domain_id'],
                   'os_password': self.creds['password'],
                   'os_project_name': self.creds['project_name'],
                   'os_project_id': '',
                   'os_project_domain_id': self.creds['project_domain_id'],
                   'os_auth_url': auth_url}

        self.conf['auth_opts'].setdefault('options', {}).update(options)

    def config(self, group=None, **kw):
        """Override some configuration values.

        The keyword arguments are the names of configuration options to
        override and their values.

        If a group argument is supplied, the overrides are applied to
        the specified configuration option group.
        """
        parent = (group and self.conf.setdefault(group, {}) or
                  self.conf)
        parent.update(kw)



# Copyright (c) 2013 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Errors below must be used to translate transport specific
errors to Zaqar errors. For example, HTTP 404s should be
raised as `ResourceNotFound`
"""

from zaqarclient import errors

__all__ = ['TransportError', 'ResourceNotFound', 'MalformedRequest',
           'UnauthorizedError', 'ForbiddenError', 'ServiceUnavailableError',
           'InternalServerError', 'ConflictError']


class TransportError(errors.ZaqarError):
    """Base class for all transport errors."""

    code = None

    def __init__(self, title=None, description=None, text=None):
        msg = 'Error response from Zaqar. Code: {0}.'.format(self.code)
        if title:
            msg += ' Title: {0}.'.format(title)
        if description:
            msg += ' Description: {0}.'.format(description)
        if text:
            msg += ' Text: {0}.'.format(text)
        super(TransportError, self).__init__(msg)


class ResourceNotFound(TransportError):
    """Indicates that a resource is missing

    This error maps to HTTP's 404
    """

    code = 404


class MalformedRequest(TransportError):
    """Indicates that a request is malformed

    This error maps to HTTP's 400
    """

    code = 400


class UnauthorizedError(TransportError):
    """Indicates that a request was not authenticated

    This error maps to HTTP's 401
    """

    code = 401


class ForbiddenError(TransportError):
    """Indicates that a request is forbidden to access the particular resource

    This error maps to HTTP's 403
    """

    code = 403


class InternalServerError(TransportError):
    """Indicates that the server encountered an unexpected situation

    This error maps to HTTP's 500
    """

    code = 500


class ServiceUnavailableError(TransportError):
    """Indicates that the server was unable to service the request

    This error maps to HTTP's 503
    """

    code = 503


class ConflictError(TransportError):
    """Indicates that the server was unable to service the request

    This error maps to HTTP's 409
    """

    code = 409



# Copyright (c) 2013  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import mock

from keystoneauth1 import session

from zaqarclient import auth
from zaqarclient.tests import base
from zaqarclient.transport import request


class TestKeystoneAuth(base.TestBase):

    def setUp(self):
        super(TestKeystoneAuth, self).setUp()

        self.auth = auth.get_backend(options=self.conf)

    @mock.patch('keystoneauth1.session.Session.get_token',
                return_value='fake-token')
    def test_no_token(self, fake_session):
        test_endpoint = 'http://example.org:8888'
        keystone_session = session.Session()

        with mock.patch.object(self.auth, '_get_endpoint') as get_endpoint:
            with mock.patch.object(self.auth,
                                   '_get_keystone_session') as get_session:

                get_endpoint.return_value = test_endpoint
                get_session.return_value = keystone_session

                req = self.auth.authenticate(1, request.Request())
                self.assertEqual(test_endpoint, req.endpoint)
                self.assertIn('X-Auth-Token', req.headers)
                self.assertIn(req.headers['X-Auth-Token'], 'fake-token')

    def test_with_token(self):
        self.auth.conf.update({"auth_token": "test-token"})
        req = request.Request(endpoint='http://example.org:8888')
        req = self.auth.authenticate(1, req)
        self.assertIn('X-Auth-Token', req.headers)
        self.assertIn(req.headers['X-Auth-Token'], 'test-token')



# Copyright (c) 2015 Catalyst IT Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tests.unit.cli import fakes

from zaqarclient.queues.v1 import cli as v1_cli
from zaqarclient.queues.v1 import iterator
from zaqarclient.queues.v1 import queues as v1_api_queues


class TestQueues(fakes.TestMessaging):
    def setUp(self):
        super(TestQueues, self).setUp()


class TestV1ListQueues(TestQueues):
    def setUp(self):
        super(TestV1ListQueues, self).setUp()
        queues_list = iterator._Iterator(self, [{'name': 'fake_queue'}],
                                         'queues',
                                         v1_api_queues.create_object(self))
        self.app.client_manager.messaging.queues.return_value = queues_list

        # Command to test
        self.cmd = v1_cli.ListQueues(self.app, None)

    def test_queues_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ('Name',)
        self.assertEqual(expected_columns, columns)
        # Check that data is correct
        expected_data = [('fake_queue',)]
        self.assertEqual(expected_data, list(data))


class TestV1CreateQueue(TestQueues):

    def setUp(self):
        super(TestV1CreateQueue, self).setUp()

        # Command to test
        self.cmd = v1_cli.CreateQueue(self.app, None)

    def test_queue_create(self):
        arglist = ['fake_queue']
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.messaging_client.queue.assert_called_with('fake_queue',
                                                       force_create=True)



# Copyright (c) 2013 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from zaqarclient._i18n import _  # noqa
from zaqarclient import errors
from zaqarclient.queues.v1 import claim as claim_api
from zaqarclient.queues.v1 import core
from zaqarclient.queues.v1 import iterator
from zaqarclient.queues.v1 import message

# NOTE(wanghao): This is copied from Zaqar server side, so if server have
# updated it someday, we should update it here to keep consistent.
QUEUE_NAME_REGEX = re.compile(r'^[a-zA-Z0-9_\-]+$')


class Queue(object):

    message_module = message

    def __init__(self, client, name, href=None, metadata=None,
                 auto_create=True, force_create=False):
        """Initialize queue object

        :param client: The client object of Zaqar.
        :type client: `object`
        :param name: Name of the queue.
        :type name: `six.string_type`
        :param href : Hypertext Reference.
        :type href:  `six.string_type`
        :param metadata : A metadata object of the queue.
        :type metadata: `dict`
        :param auto_create: If create the queue automatically in database.
        :type auto_create: `boolean`
        :param force_create: If create the queue and skip the API version
            check, which is useful for command line interface.
        :type force_create: `boolean`
        :returns: The queue object.
        """
        self.client = client

        if name == "":
            raise ValueError(_('Queue name does not have a value'))

        if not QUEUE_NAME_REGEX.match(str(name)):
            raise ValueError(_('The queue name may only contain ASCII '
                               'letters, digits, underscores and dashes.'))

        # NOTE(flaper87) Queue Info
        self._name = name
        self._metadata = metadata
        self._href = href

        # NOTE(flwang): If force_create is True, then even though auto_create
        # is not True, the queue should be created anyway.
        if auto_create or force_create:
            self.ensure_exists(force_create=force_create)

    @property
    def name(self):
        return self._name

    @property
    def href(self):
        return self._href

    @property
    def metadata_dict(self):
        return dict(self.metadata())

    def exists(self):
        """Checks if the queue exists."""
        req, trans = self.client._request_and_transport()
        if self.client.api_version >= 1.1:
            raise errors.InvalidOperation("Unavailable on versions >= 1.1")
        else:
            return core.queue_exists(trans, req, self._name)

    def ensure_exists(self, force_create=False):
        """Ensures a queue exists

        This method is not race safe,
        the queue could've been deleted
        right after it was called.
        """
        req, trans = self.client._request_and_transport()
        if force_create or self.client.api_version < 1.1:
            core.queue_create(trans, req, self._name)

    def metadata(self, new_meta=None, force_reload=False):
        """Get metadata and return it

        :param new_meta: A dictionary containing
            an updated metadata object. If present
            the queue metadata will be updated in
            remote server. If the new_meta is empty,
            the metadata object will be cleared.
        :type new_meta: `dict`
        :param force_reload: Whether to ignored the
            cached metadata and reload it from the
            server.
        :type force_reload: `bool`

        :returns: The queue metadata.
        """
        req, trans = self.client._request_and_transport()

        # NOTE(jeffrey4l): Ensure that metadata is cleared when the new_meta
        # is an empty dict.
        if new_meta is not None:
            if self.client.api_version == 1.1:
                raise RuntimeError("V1.1 doesn't support to set the queue's "
                                   "metadata. Please use V1.0 or V2.")
            core.queue_set_metadata(trans, req, self._name, new_meta)
            self._metadata = new_meta

        # TODO(flaper87): Cache with timeout
        if self._metadata and not force_reload:
            return self._metadata

        if self.client.api_version >= 1.1:
            self._metadata = core.queue_get(trans, req, self._name)
        else:
            self._metadata = core.queue_get_metadata(trans, req, self._name)
        return self._metadata

    @property
    def stats(self):
        req, trans = self.client._request_and_transport()
        return core.queue_get_stats(trans, req, self._name)

    def delete(self):
        req, trans = self.client._request_and_transport()
        core.queue_delete(trans, req, self._name)

    # Messages API

    def post(self, messages):
        """Posts one or more messages to this queue

        :param messages: One or more messages to post
        :type messages: `list` or `dict`

        :returns: A dict with the result of this operation.
        :rtype: `dict`
        """
        if not isinstance(messages, list):
            messages = [messages]

        if self.client.api_version >= 1.1:
            messages = {'messages': messages}

        req, trans = self.client._request_and_transport()

        # TODO(flaper87): Return a list of messages
        return core.message_post(trans, req,
                                 self._name, messages)

    def message(self, message_id):
        """Gets a message by id

        :param message_id: Message's reference
        :type message_id: `six.text_type`

        :returns: A message
        :rtype: `dict`
        """
        req, trans = self.client._request_and_transport()
        msg = core.message_get(trans, req, self._name,
                               message_id)
        return self.message_module.Message(self, **msg)

    def messages(self, *messages, **params):
        """Gets a list of messages from the server

        This method returns a list of messages, it can be
        used to retrieve a set of messages by id or to
        walk through the active messages by using the
        collection endpoint.

        The `messages` and `params` params are mutually exclusive
        and the former has the priority.

        :param messages: List of messages' ids to retrieve.
        :type messages: *args of `six.string_type`

        :param params: Filters to use for getting messages
        :type params: **kwargs dict.

        :returns: List of messages
        :rtype: `list`
        """
        req, trans = self.client._request_and_transport()

        # TODO(flaper87): Return a MessageIterator.
        # This iterator should handle limits, pagination
        # and messages deserialization.

        if messages:
            msgs = core.message_get_many(trans, req,
                                         self._name, messages)
        else:
            # NOTE(flaper87): It's safe to access messages
            # directly. If something wrong happens, the core
            # API will raise the right exceptions.
            msgs = core.message_list(trans, req,
                                     self._name,
                                     **params)

        return iterator._Iterator(self.client,
                                  msgs,
                                  'messages',
                                  self.message_module.create_object(self))

    def delete_messages(self, *messages):
        """Deletes a set of messages from the server

        :param messages: List of messages' ids to delete.
        :type messages: *args of `six.string_type`
        """

        req, trans = self.client._request_and_transport()
        return core.message_delete_many(trans, req, self._name,
                                        set(messages))

    def pop(self, count=1):
        """Pop `count` messages from the server

        :param count: Number of messages to pop.
        :type count: int

        :returns: List of messages
        :rtype: `list`
        """

        req, trans = self.client._request_and_transport()
        msgs = core.message_pop(trans, req, self._name, count=count)
        return iterator._Iterator(self.client,
                                  msgs,
                                  'messages',
                                  self.message_module.create_object(self))

    def claim(self, id=None, ttl=None, grace=None,
              limit=None):
        return claim_api.Claim(self, id=id, ttl=ttl, grace=grace, limit=limit)


def create_object(parent):
    return lambda args: Queue(parent, args["name"], auto_create=False)



# Copyright (c) 2013 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from distutils import version
import json

from oslo_utils import importutils

from zaqarclient.common import http
from zaqarclient.transport import base
from zaqarclient.transport import response

osprofiler_web = importutils.try_import("osprofiler.web")


class HttpTransport(base.Transport):

    def __init__(self, options):
        super(HttpTransport, self).__init__(options)
        self.client = http.Client()

    def _prepare(self, request):
        if not request.api:
            return request.endpoint, 'GET', request

        # TODO(flaper87): Validate if the user
        # explicitly wants so. Validation must
        # happen before any other operation here.
        # request.validate()

        schema = {}
        ref_params = {}
        ref = request.ref

        if request.operation:
            schema = request.api.get_schema(request.operation)
            ref = ref or schema.get('ref', '')

        # FIXME(flaper87): We expect the endpoint
        # to have the API version label already,
        # however in a follow-your-nose implementation
        # it should be the other way around.
        ref = ref.lstrip('/' + request.api.label)

        for param in list(request.params.keys()):
            if '{{{0}}}'.format(param) in ref:
                value = request.params.pop(param)

                # NOTE(flaper87): Zaqar API parses
                # sequences encoded as '1,2,3,4'. Let's
                # encode lists, tuples and sets before
                # sending them to the server.
                if isinstance(value, (list, tuple, set)):
                    value = ','.join(value)

                ref_params[param] = value

        url = '{0}/{1}/{2}'.format(request.endpoint.rstrip('/'),
                                   request.api.label,
                                   ref.format(**ref_params))
        return url, schema.get('method', 'GET'), request

    def send(self, request):
        url, method, request = self._prepare(request)

        # NOTE(flape87): Do not modify
        # request's headers directly.
        headers = request.headers.copy()
        if (request.operation == 'queue_update' and
                (version.LooseVersion(request.api.label) >=
                    version.LooseVersion('v2'))):
            headers['content-type'] = \
                'application/openstack-messaging-v2.0-json-patch'
        else:
            headers['content-type'] = 'application/json'

        if osprofiler_web:
            headers.update(osprofiler_web.get_trace_id_headers())

        if request.verify:
            if request.cert:
                verify = request.cert
            else:
                verify = True
        else:
            verify = False
        resp = self.client.request(method,
                                   url=url,
                                   params=request.params,
                                   headers=headers,
                                   data=request.content,
                                   verify=verify)

        if resp.status_code in self.http_to_zaqar:
            kwargs = {}
            try:
                error_body = json.loads(resp.text)
                kwargs['title'] = error_body['title']
                kwargs['description'] = error_body['description']
            except Exception:
                # TODO(flaper87): Log this exception
                # but don't stop raising the corresponding
                # exception
                # Note(Eva-i): most of the error responses from Zaqar have
                # dict with title and description in their bodies. If it's not
                # the case, let's just show body text.
                kwargs['text'] = resp.text
            raise self.http_to_zaqar[resp.status_code](**kwargs)

        # NOTE(flaper87): This reads the whole content
        # and will consume any attempt of streaming.
        return response.Response(request, resp.text,
                                 headers=resp.headers,
                                 status_code=resp.status_code)
